//***************************************************************************
//
// ARCBALL.H
// by Harlen C. Batagelo
//
//***************************************************************************

#ifndef _ARCBALL_H
#define _ARCBALL_H

/////////////////////////////////////////////////////////////////////////////
// QUATERNION type
/////////////////////////////////////////////////////////////////////////////
typedef struct QUATERNION
{
public:
    QUATERNION() {}
    QUATERNION( const float * );
    QUATERNION( float x, float y, float z, float w );

    // casting
    operator float* ();
    operator const float* () const;

    // assignment operators
    QUATERNION& operator += ( const QUATERNION& );
    QUATERNION& operator -= ( const QUATERNION& );
    QUATERNION& operator *= ( const QUATERNION& );
    QUATERNION& operator *= ( float );
    QUATERNION& operator /= ( float );

    // unary operators
    QUATERNION  operator + () const;
    QUATERNION  operator - () const;

    // binary operators
    QUATERNION operator + ( const QUATERNION& ) const;
    QUATERNION operator - ( const QUATERNION& ) const;
    QUATERNION operator * ( const QUATERNION& ) const;
    QUATERNION operator * ( float ) const;
    QUATERNION operator / ( float ) const;

    friend QUATERNION operator * (float, const QUATERNION& );

    bool operator == ( const QUATERNION& ) const;
    bool operator != ( const QUATERNION& ) const;

    float x, y, z, w;
} QUATERNION, *LPQUATERNION;


/////////////////////////////////////////////////////////////////////////////
// VECTOR type
/////////////////////////////////////////////////////////////////////////////
typedef struct _VECTOR {
    union { float x; float dvX; };
    union { float y; float dvY; };
    union { float z; float dvZ; };

public:

    // Constructors
    _VECTOR() { }
    _VECTOR(float f);
    _VECTOR(float _x, float _y, float _z);
    _VECTOR(const float f[3]);

    // Access grants
    const float&operator[](int i) const;
    float&operator[](int i);

    // Assignment operators
    _VECTOR& operator += (const _VECTOR& v);
    _VECTOR& operator -= (const _VECTOR& v);
    _VECTOR& operator *= (const _VECTOR& v);
    _VECTOR& operator /= (const _VECTOR& v);
    _VECTOR& operator *= (float s);
    _VECTOR& operator /= (float s);

    // Unary operators
    friend _VECTOR operator + (const _VECTOR& v);
    friend _VECTOR operator - (const _VECTOR& v);

    // Binary operators

    // Addition and subtraction
    friend _VECTOR operator + (const _VECTOR& v1, const _VECTOR& v2);
    friend _VECTOR operator - (const _VECTOR& v1, const _VECTOR& v2);
    // Scalar multiplication and division
    friend _VECTOR operator * (const _VECTOR& v, float s);
	friend _VECTOR operator * (float s, const _VECTOR& v);
	friend _VECTOR operator / (const _VECTOR& v, float s);
    // Memberwise multiplication and division
	friend _VECTOR operator * (const _VECTOR& v1, const _VECTOR& v2);
	friend _VECTOR operator / (const _VECTOR& v1, const _VECTOR& v2);

    // Vector dominance
	friend int operator < (const _VECTOR& v1, const _VECTOR& v2);
	friend int operator <= (const _VECTOR& v1, const _VECTOR& v2);

    // Bitwise equality
	friend int operator == (const _VECTOR& v1, const _VECTOR& v2);

    // Length-related functions
	friend float SquareMagnitude (const _VECTOR& v);
	friend float Magnitude (const _VECTOR& v);

    // Returns vector with same direction and unit length
	friend _VECTOR Normalize (const _VECTOR& v);

    // Return min/max component of the input vector
	friend float Min (const _VECTOR& v);
	friend float Max (const _VECTOR& v);

    // Return memberwise min/max of input vectors
	friend _VECTOR Minimize (const _VECTOR& v1, const _VECTOR& v2);
	friend _VECTOR Maximize (const _VECTOR& v1, const _VECTOR& v2);

    // Dot and cross product
	friend float DotProduct (const _VECTOR& v1, const _VECTOR& v2);
	friend _VECTOR CrossProduct (const _VECTOR& v1, const _VECTOR& v2);

} VECTOR, *LPVECTOR;

/////////////////////////////////////////////////////////////////////////////
// VECTOR3 type
/////////////////////////////////////////////////////////////////////////////
typedef struct VECTOR3 : public VECTOR {
public:
    VECTOR3() {};
    VECTOR3( const float * );
    VECTOR3( const VECTOR& );
    VECTOR3( float x, float y, float z );

    // casting
    operator float* ();
    operator const float* () const;

    // assignment operators
    VECTOR3& operator += ( const VECTOR3& );
    VECTOR3& operator -= ( const VECTOR3& );
    VECTOR3& operator *= ( float );
    VECTOR3& operator /= ( float );

    // unary operators
    VECTOR3 operator + () const;
    VECTOR3 operator - () const;

    // binary operators
    VECTOR3 operator + ( const VECTOR3& ) const;
    VECTOR3 operator - ( const VECTOR3& ) const;
    VECTOR3 operator * ( float ) const;
    VECTOR3 operator / ( float ) const;

    friend VECTOR3 operator * ( float, const struct VECTOR3& );

    bool operator == ( const VECTOR3& ) const;
    bool operator != ( const VECTOR3& ) const;

} VECTOR3, *LPVECTOR3;

/////////////////////////////////////////////////////////////////////////////
// POLYGON type
/////////////////////////////////////////////////////////////////////////////
typedef struct POLYGON {
public:
	int iEdgeTotal;	  // Number of edges
	int iEdgeIndices[32]; // Vertex indices 
						  // Each edge uses two array elements (max 16 edges)

    POLYGON() { iEdgeTotal = 0; iEdgeIndices[0] = -1; }

	void addIdx (int iIdx) { iEdgeIndices[iEdgeTotal++] = iIdx; }
} POLYGON, *LPPOLYGON;

/////////////////////////////////////////////////////////////////////////////
// MATRIX type
/////////////////////////////////////////////////////////////////////////////
typedef struct _MATRIX {
	union 
	{
		struct 
		{
			float _11, _12, _13, _14;
			float _21, _22, _23, _24;
			float _31, _32, _33, _34;
			float _41, _42, _43, _44;
		};
		float m[4][4];
	};

    _MATRIX() { }
    _MATRIX( float _m00, float _m01, float _m02, float _m03,
             float _m10, float _m11, float _m12, float _m13,
             float _m20, float _m21, float _m22, float _m23,
             float _m30, float _m31, float _m32, float _m33 )
	{
		m[0][0] = _m00; m[0][1] = _m01; m[0][2] = _m02; m[0][3] = _m03;
		m[1][0] = _m10; m[1][1] = _m11; m[1][2] = _m12; m[1][3] = _m13;
		m[2][0] = _m20; m[2][1] = _m21; m[2][2] = _m22; m[2][3] = _m23;
		m[3][0] = _m30; m[3][1] = _m31; m[3][2] = _m32; m[3][3] = _m33;
	}

    float& operator()(int iRow, int iColumn) { return m[iRow][iColumn]; }
    const float& operator()(int iRow, int iColumn) const { return m[iRow][iColumn]; }
//    friend _MATRIX operator* (const _MATRIX&, const _MATRIX&);
} MATRIX, *LPMATRIX;

/////////////////////////////////////////////////////////////////////////////
// XMATRIX type
/////////////////////////////////////////////////////////////////////////////
typedef struct XMATRIX : public MATRIX {
public:
    XMATRIX() {};
    XMATRIX( const float * );
    XMATRIX( const MATRIX& );
    XMATRIX( float _11, float _12, float _13, float _14,
             float _21, float _22, float _23, float _24,
             float _31, float _32, float _33, float _34,
             float _41, float _42, float _43, float _44 );

    // access grants
    float& operator () ( unsigned int Row, unsigned int Col );
    float  operator () ( unsigned int Row, unsigned int Col ) const;

    // casting operators
    operator float* ();
    operator const float* () const;

    // assignment operators
    XMATRIX& operator *= ( const XMATRIX& );
    XMATRIX& operator += ( const XMATRIX& );
    XMATRIX& operator -= ( const XMATRIX& );
    XMATRIX& operator *= ( float );
    XMATRIX& operator /= ( float );

    // unary operators
    XMATRIX operator + () const;
    XMATRIX operator - () const;

    // binary operators
    XMATRIX operator * ( const XMATRIX& ) const;
    XMATRIX operator + ( const XMATRIX& ) const;
    XMATRIX operator - ( const XMATRIX& ) const;
    XMATRIX operator * ( float ) const;
    XMATRIX operator / ( float ) const;

    friend XMATRIX operator * ( float, const XMATRIX& );

    bool operator == ( const XMATRIX& ) const;
    bool operator != ( const XMATRIX& ) const;

} XMATRIX, *LPXMATRIX;

/////////////////////////////////////////////////////////////////////////////
// ArcBall class
/////////////////////////////////////////////////////////////////////////////
class CArcBall
{
	int        m_iWidth;   // ArcBall's window width
	int        m_iHeight;  // ArcBall's window height
	float      m_fRadius;  // ArcBall's radius in screen coords
	float      m_fRadius2; // ArcBall's radius in screen coords

	QUATERNION m_qDown;               // Quaternion before button down
	QUATERNION m_qNow;                // Composite quaternion for current drag
	XMATRIX    m_matRotation;         // Matrix for arcball's orientation
	XMATRIX    m_matRotationDelta;    // Matrix for arcball's orientation
	XMATRIX    m_matTranslation;      // Matrix for arcball's position
	XMATRIX    m_matTranslationDelta; // Matrix for arcball's position
	bool       m_bDrag;               // Whether user is dragging arcball

    int	   iCurMouseX;		// Saved mouse position
    int    iCurMouseY;
    VECTOR3 m_vDown;		// Button down quaternion
    VECTOR3 m_vCur;			// Current quaternion
    int iMouseX, iMouseY;	// Current mouse position

public:
	void LeftButtonDown (int iMouseX, int iMouseY);
	void LeftButtonUp ();
	void LeftButtonDownMove (int iMouseX, int iMouseY);

	VECTOR3    ScreenToVector( int sx, int sy );

	XMATRIX*	GetRotationMatrix()         { return &m_matRotation; }
	XMATRIX*	GetRotationDeltaMatrix()    { return &m_matRotationDelta; }
	void		GetTranslationMatrix(float *m)      
	{
		int ct = 0;

		for (int i = 0; i < 4; i++)
		{
			for (int j = 0; j < 4; j++)
			{
				m[ct++] = m_matTranslation(i,j);
			}
		}
	}
	XMATRIX*	GetTranslationDeltaMatrix() { return &m_matTranslationDelta; }
	bool        IsBeingDragged()            { return m_bDrag; }

	void        SetRadius( float fRadius );
	void        SetWindow( int w, int h, float r=0.9 );

	CArcBall();
};

#include "arcball.inl"

#endif