//***************************************************************************
//
// ARCBALL.INL
// by Harlen C. Batagelo
//
//***************************************************************************

#ifndef __ARCBALL_INL__
#define __ARCBALL_INL__

#include <memory.h>
#include <math.h>

/////////////////////////////////////////////////////////////////////////////
// Inline Class Methods
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Vector
/////////////////////////////////////////////////////////////////////////////

// Constructors
inline _VECTOR::_VECTOR(float f)
{
    x = y = z = f;
}

inline _VECTOR::_VECTOR(float _x, float _y, float _z)
{
    x = _x; y = _y; z = _z;
}

inline _VECTOR::_VECTOR(const float f[3])
{
    x = f[0]; y = f[1]; z = f[2];
}

// Access grants
inline const float& _VECTOR::operator[](int i) const
{
    return (&x)[i];
}

inline float& _VECTOR::operator[](int i)
{
    return (&x)[i];
}

// Assignment operators
inline _VECTOR& _VECTOR::operator += (const _VECTOR& v)
{
   x += v.x;   y += v.y;   z += v.z;
   return *this;
}

inline _VECTOR& _VECTOR::operator -= (const _VECTOR& v)
{
   x -= v.x;   y -= v.y;   z -= v.z;
   return *this;
}

inline _VECTOR& _VECTOR::operator *= (const _VECTOR& v)
{
   x *= v.x;   y *= v.y;   z *= v.z;
   return *this;
}

inline _VECTOR& _VECTOR::operator /= (const _VECTOR& v)
{
   x /= v.x;   y /= v.y;   z /= v.z;
   return *this;
}

inline _VECTOR& _VECTOR::operator *= (float s)
{
   x *= s;   y *= s;   z *= s;
   return *this;
}

inline _VECTOR& _VECTOR::operator /= (float s)
{
   x /= s;   y /= s;   z /= s;
   return *this;
}

inline _VECTOR operator + (const _VECTOR& v)
{
   return v;
}

inline _VECTOR operator - (const _VECTOR& v)
{
   return _VECTOR(-v.x, -v.y, -v.z);
}

inline _VECTOR operator + (const _VECTOR& v1, const _VECTOR& v2)
{
   return _VECTOR(v1.x+v2.x, v1.y+v2.y, v1.z+v2.z);
}

inline _VECTOR operator - (const _VECTOR& v1, const _VECTOR& v2)
{
   return _VECTOR(v1.x-v2.x, v1.y-v2.y, v1.z-v2.z);
}

inline _VECTOR operator * (const _VECTOR& v1, const _VECTOR& v2)
{
   return _VECTOR(v1.x*v2.x, v1.y*v2.y, v1.z*v2.z);
}

inline _VECTOR operator / (const _VECTOR& v1, const _VECTOR& v2)
{
   return _VECTOR(v1.x/v2.x, v1.y/v2.y, v1.z/v2.z);
}

inline int operator < (const _VECTOR& v1, const _VECTOR& v2)
{
   return v1[0] < v2[0] && v1[1] < v2[1] && v1[2] < v2[2];
}

inline int operator <= (const _VECTOR& v1, const _VECTOR& v2)
{
   return v1[0] <= v2[0] && v1[1] <= v2[1] && v1[2] <= v2[2];
}

inline _VECTOR operator * (const _VECTOR& v, float s)
{
   return _VECTOR(s*v.x, s*v.y, s*v.z);
}

inline _VECTOR operator * (float s, const _VECTOR& v)
{
   return _VECTOR(s*v.x, s*v.y, s*v.z);
}

inline _VECTOR operator / (const _VECTOR& v, float s)
{
   return _VECTOR(v.x/s, v.y/s, v.z/s);
}

inline int operator == (const _VECTOR& v1, const _VECTOR& v2)
{
   return v1.x==v2.x && v1.y==v2.y && v1.z == v2.z;
}

inline float Magnitude (const _VECTOR& v)
{
   return (float) sqrt(SquareMagnitude(v));
}

inline float SquareMagnitude (const _VECTOR& v)
{
   return v.x*v.x + v.y*v.y + v.z*v.z;
}

inline _VECTOR Normalize (const _VECTOR& v)
{
   return v / Magnitude(v);
}

inline float Min (const _VECTOR& v)
{
   float ret = v.x;
   if (v.y < ret) ret = v.y;
   if (v.z < ret) ret = v.z;
   return ret;
}

inline float Max (const _VECTOR& v)
{
   float ret = v.x;
   if (ret < v.y) ret = v.y;
   if (ret < v.z) ret = v.z;
   return ret;
}

inline _VECTOR Minimize (const _VECTOR& v1, const _VECTOR& v2)
{
   return _VECTOR( v1[0] < v2[0] ? v1[0] : v2[0],
                   v1[1] < v2[1] ? v1[1] : v2[1],
                   v1[2] < v2[2] ? v1[2] : v2[2]);
}

inline _VECTOR Maximize (const _VECTOR& v1, const _VECTOR& v2)
{
   return _VECTOR( v1[0] > v2[0] ? v1[0] : v2[0],
                   v1[1] > v2[1] ? v1[1] : v2[1],
                   v1[2] > v2[2] ? v1[2] : v2[2]);
}

inline float DotProduct (const _VECTOR& v1, const _VECTOR& v2)
{
   return v1.x*v2.x + v1.y * v2.y + v1.z*v2.z;
}

inline _VECTOR CrossProduct (const _VECTOR& v1, const _VECTOR& v2)
{
	_VECTOR result;

	result[0] = v1[1] * v2[2] - v1[2] * v2[1];
	result[1] = v1[2] * v2[0] - v1[0] * v2[2];
	result[2] = v1[0] * v2[1] - v1[1] * v2[0];

	return result;
}

inline _MATRIX operator* (const _MATRIX& a, const _MATRIX& b)
{
    _MATRIX ret;
    for (int i=0; i<4; i++) {
        for (int j=0; j<4; j++) {
            ret(i, j) = 0.0f;
            for (int k=0; k<4; k++) {
                ret(i, j) += a(i, k) * b(k, j);
            }
        }
    }
    return ret;
}

/////////////////////////////////////////////////////////////////////////////
// 3D Vector
/////////////////////////////////////////////////////////////////////////////
inline VECTOR3::VECTOR3( const float *pf )
{
	x = pf[0];
	y = pf[1];
	z = pf[2];
}

inline VECTOR3::VECTOR3( const VECTOR& v )
{
	x = v.x;
	y = v.y;
	z = v.z;
}

inline VECTOR3::VECTOR3( float fx, float fy, float fz )
{
	x = fx;
	y = fy;
	z = fz;
}

// casting
inline VECTOR3::operator float* ()
{
    return (float *) &x;
}

inline VECTOR3::operator const float* () const
{
    return (const float *) &x;
}

// assignment operators
inline VECTOR3& VECTOR3::operator += ( const VECTOR3& v )
{
	x += v.x;
	y += v.y;
	z += v.z;
	return *this;
}

inline VECTOR3& VECTOR3::operator -= ( const VECTOR3& v )
{
	x -= v.x;
	y -= v.y;
	z -= v.z;
	return *this;
}

inline VECTOR3& VECTOR3::operator *= ( float f )
{
	x *= f;
	y *= f;
	z *= f;
	return *this;
}

inline VECTOR3& VECTOR3::operator /= ( float f )
{
	float fInv = 1.0f / f;
	x *= fInv;
	y *= fInv;
	z *= fInv;
	return *this;
}

// unary operators
inline VECTOR3 VECTOR3::operator + () const
{
    return *this;
}

inline VECTOR3 VECTOR3::operator - () const
{
    return VECTOR3(-x, -y, -z);
}

// binary operators
inline VECTOR3 VECTOR3::operator + ( const VECTOR3& v ) const
{
    return VECTOR3(x + v.x, y + v.y, z + v.z);
}

inline VECTOR3 VECTOR3::operator - ( const VECTOR3& v ) const
{
    return VECTOR3(x - v.x, y - v.y, z - v.z);
}

inline VECTOR3 VECTOR3::operator * ( float f ) const
{
    return VECTOR3(x * f, y * f, z * f);
}

inline VECTOR3 VECTOR3::operator / ( float f ) const
{
    float fInv = 1.0f / f;
    return VECTOR3(x * fInv, y * fInv, z * fInv);
}

inline VECTOR3 operator * ( float f, const struct VECTOR3& v )
{
    return VECTOR3(f * v.x, f * v.y, f * v.z);
}

inline bool VECTOR3::operator == ( const VECTOR3& v ) const
{
    return x == v.x && y == v.y && z == v.z;
}

inline bool VECTOR3::operator != ( const VECTOR3& v ) const
{
    return x != v.x || y != v.y || z != v.z;
}

inline XMATRIX ZeroMatrix()
{
	XMATRIX pRes;

	for (int i = 0; i < 4; i++)
	{
		for (int j = 0; j < 4; j++)
		{
			pRes(i,j) = 0.0;
		}
	}
	
	return pRes;
}

// Matrix multiplication.  The result represents the transformation M2
// followed by the transformation M1.  (Out = M1 * M2)
inline XMATRIX* MatrixMultiply ( XMATRIX *pOut, const XMATRIX *pM1, 
								 const XMATRIX *pM2 )
{
	XMATRIX pRes = ZeroMatrix();

	for (int i = 0; i < 4; i++)
	{
		for (int j = 0; j < 4; j++)
		{
			for (int k = 0; k < 4; k++)
			{
				pRes(i,j) += pM1->m[i][k] * pM2->m[k][j];
			}
		}
	}

	*pOut = pRes;

	return pOut;
}

inline XMATRIX* MatrixTranspose ( XMATRIX *pOut, const XMATRIX *pM )
{
	XMATRIX pRes;

	for (int i = 0; i < 4; i++)
	{
		for (int j = 0; j < 4; j++)
		{
			pRes(j,i) = pM->m[i][j];
		}
	}

	*pOut = pRes;

	return pOut;
}

inline void MatrixInverse (float *pOut, XMATRIX *pIn)
{
	XMATRIX pInv = *pIn;

	// Transpose 3x3 part
	float a12 = pInv._12;
	float a13 = pInv._13;
	float a23 = pInv._23;
	pInv._12 = pInv._21;
	pInv._13 = pInv._31;
	pInv._23 = pInv._32;
	pInv._21 = a12;
	pInv._31 = a13;
	pInv._32 = a23;

	// Negate the translation part and multiply by the reciprocal of the rotation part
 	float a41 = (pInv._41 * pInv._11) + (pInv._42 * pInv._21) + (pInv._43 * pInv._31);
	float a42 = (pInv._41 * pInv._12) + (pInv._42 * pInv._22) + (pInv._43 * pInv._32);
	float a43 = (pInv._41 * pInv._13) + (pInv._42 * pInv._23) + (pInv._43 * pInv._33);

	pInv._41 = -a41;
	pInv._42 = -a42;
	pInv._43 = -a43;

	pOut[ 0] = pInv._11;
	pOut[ 1] = pInv._21;
	pOut[ 2] = pInv._31;
	pOut[ 3] = pInv._41;

	pOut[ 4] = pInv._12;
	pOut[ 5] = pInv._22;
	pOut[ 6] = pInv._32;
	pOut[ 7] = pInv._42;

	pOut[ 8] = pInv._13;
	pOut[ 9] = pInv._23;
	pOut[10] = pInv._33;
	pOut[11] = pInv._43;

	pOut[12] = pInv._14;
	pOut[13] = pInv._24;
	pOut[14] = pInv._34;
	pOut[15] = pInv._44;
}

// Quaternion multiplication.  The result represents the rotation Q2 
// followed by the rotation Q1.  (Out = Q2 * Q1)
inline QUATERNION* QuaternionMultiply ( QUATERNION *pOut, const QUATERNION *pQ1, 
										const QUATERNION *pQ2 )
{
	float s1 = pQ1->w;
	float s2 = pQ2->w;

	VECTOR3 v1(pQ1->x, pQ1->y, pQ1->z);
	VECTOR3 v2(pQ2->x, pQ2->y, pQ2->z);

	pOut->w = s2*s1 - DotProduct(v2, v1);
	VECTOR3 res = (s2*v1) + (s1*v2) + (VECTOR3)(CrossProduct(v2, v1));
	pOut->x = res.x;
	pOut->y = res.y;
	pOut->z = res.z;

	return pOut;
}

#define POW2(x) ((x)*(x))

// Build a matrix from a quaternion
inline XMATRIX*  MatrixRotationQuaternion (XMATRIX *pOut, const QUATERNION *pQ)
{
	pOut->m[0][0] = POW2(pQ->w)+POW2(pQ->x)-POW2(pQ->y)-POW2(pQ->z);
	pOut->m[0][1] = 2*pQ->x*pQ->y + 2*pQ->w*pQ->z;
	pOut->m[0][2] = 2*pQ->x*pQ->z - 2*pQ->w*pQ->y;
	pOut->m[0][3] = 0;

	pOut->m[1][0] = 2*pQ->x*pQ->y - 2*pQ->w*pQ->z;
	pOut->m[1][1] = POW2(pQ->w)-POW2(pQ->x)+POW2(pQ->y)-POW2(pQ->z);
	pOut->m[1][2] = 2*pQ->y*pQ->z + 2*pQ->w*pQ->x;
	pOut->m[1][3] = 0;

	pOut->m[2][0] = 2*pQ->x*pQ->z + 2*pQ->w*pQ->y;
	pOut->m[2][1] = 2*pQ->y*pQ->z - 2*pQ->w*pQ->x;
	pOut->m[2][2] = POW2(pQ->w)-POW2(pQ->x)-POW2(pQ->y)+POW2(pQ->z);
	pOut->m[2][3] = 0;

	pOut->m[3][0] = 0;
	pOut->m[3][1] = 0;
	pOut->m[3][2] = 0;
	pOut->m[3][3] = POW2(pQ->w)+POW2(pQ->x)+POW2(pQ->y)+POW2(pQ->z);

	return pOut;
}

/////////////////////////////////////////////////////////////////////////////
// Matrix
/////////////////////////////////////////////////////////////////////////////
inline XMATRIX::XMATRIX( const float* pf )
{
	memcpy(&_11, pf, sizeof(XMATRIX));
}

inline XMATRIX::XMATRIX( const MATRIX& mat )
{
	memcpy(&_11, &mat, sizeof(XMATRIX));
}

inline XMATRIX::XMATRIX( float f11, float f12, float f13, float f14,
                        float f21, float f22, float f23, float f24,
                        float f31, float f32, float f33, float f34,
                        float f41, float f42, float f43, float f44 )
{
	_11 = f11; _12 = f12; _13 = f13; _14 = f14;
	_21 = f21; _22 = f22; _23 = f23; _24 = f24;
	_31 = f31; _32 = f32; _33 = f33; _34 = f34;
	_41 = f41; _42 = f42; _43 = f43; _44 = f44;
}

// access grants
inline float& XMATRIX::operator () ( unsigned int iRow, unsigned int iCol )
{
	return m[iRow][iCol];
}

inline float XMATRIX::operator () ( unsigned int iRow, unsigned int iCol ) const
{
	return m[iRow][iCol];
}

// casting operators
inline XMATRIX::operator float* ()
{
    return (float *) &_11;
}

inline XMATRIX::operator const float* () const
{
    return (const float *) &_11;
}

// assignment operators
inline XMATRIX& XMATRIX::operator *= ( const XMATRIX& mat )
{
    MatrixMultiply(this, this, &mat);
    return *this;
}

inline XMATRIX& XMATRIX::operator += ( const XMATRIX& mat )
{
    _11 += mat._11; _12 += mat._12; _13 += mat._13; _14 += mat._14;
    _21 += mat._21; _22 += mat._22; _23 += mat._23; _24 += mat._24;
    _31 += mat._31; _32 += mat._32; _33 += mat._33; _34 += mat._34;
    _41 += mat._41; _42 += mat._42; _43 += mat._43; _44 += mat._44;
    return *this;
}

inline XMATRIX& XMATRIX::operator -= ( const XMATRIX& mat )
{
	_11 -= mat._11; _12 -= mat._12; _13 -= mat._13; _14 -= mat._14;
	_21 -= mat._21; _22 -= mat._22; _23 -= mat._23; _24 -= mat._24;
	_31 -= mat._31; _32 -= mat._32; _33 -= mat._33; _34 -= mat._34;
	_41 -= mat._41; _42 -= mat._42; _43 -= mat._43; _44 -= mat._44;
	return *this;
}

inline XMATRIX& XMATRIX::operator *= ( float f )
{
	_11 *= f; _12 *= f; _13 *= f; _14 *= f;
	_21 *= f; _22 *= f; _23 *= f; _24 *= f;
	_31 *= f; _32 *= f; _33 *= f; _34 *= f;
	_41 *= f; _42 *= f; _43 *= f; _44 *= f;
	return *this;
}

inline XMATRIX& XMATRIX::operator /= ( float f )
{
    float fInv = 1.0f / f;
    _11 *= fInv; _12 *= fInv; _13 *= fInv; _14 *= fInv;
    _21 *= fInv; _22 *= fInv; _23 *= fInv; _24 *= fInv;
    _31 *= fInv; _32 *= fInv; _33 *= fInv; _34 *= fInv;
    _41 *= fInv; _42 *= fInv; _43 *= fInv; _44 *= fInv;
    return *this;
}

// unary operators
inline XMATRIX XMATRIX::operator + () const
{
    return *this;
}

inline XMATRIX XMATRIX::operator - () const
{
	return XMATRIX(-_11, -_12, -_13, -_14,
                   -_21, -_22, -_23, -_24,
                   -_31, -_32, -_33, -_34,
                   -_41, -_42, -_43, -_44);
}

// binary operators
inline XMATRIX XMATRIX::operator * ( const XMATRIX& mat ) const
{
    XMATRIX matT;
    MatrixMultiply(&matT, this, &mat);
    return matT;
}

inline XMATRIX XMATRIX::operator + ( const XMATRIX& mat ) const
{
    return XMATRIX(_11 + mat._11, _12 + mat._12, _13 + mat._13, _14 + mat._14,
                   _21 + mat._21, _22 + mat._22, _23 + mat._23, _24 + mat._24,
                   _31 + mat._31, _32 + mat._32, _33 + mat._33, _34 + mat._34,
                   _41 + mat._41, _42 + mat._42, _43 + mat._43, _44 + mat._44);
}

inline XMATRIX XMATRIX::operator - ( const XMATRIX& mat ) const
{
    return XMATRIX(_11 - mat._11, _12 - mat._12, _13 - mat._13, _14 - mat._14,
                   _21 - mat._21, _22 - mat._22, _23 - mat._23, _24 - mat._24,
                   _31 - mat._31, _32 - mat._32, _33 - mat._33, _34 - mat._34,
                   _41 - mat._41, _42 - mat._42, _43 - mat._43, _44 - mat._44);
}

inline XMATRIX XMATRIX::operator * ( float f ) const
{
	return XMATRIX(_11 * f, _12 * f, _13 * f, _14 * f,
                   _21 * f, _22 * f, _23 * f, _24 * f,
                   _31 * f, _32 * f, _33 * f, _34 * f,
                   _41 * f, _42 * f, _43 * f, _44 * f);
}

inline XMATRIX XMATRIX::operator / ( float f ) const
{
    float fInv = 1.0f / f;
    return XMATRIX(_11 * fInv, _12 * fInv, _13 * fInv, _14 * fInv,
                   _21 * fInv, _22 * fInv, _23 * fInv, _24 * fInv,
                   _31 * fInv, _32 * fInv, _33 * fInv, _34 * fInv,
                   _41 * fInv, _42 * fInv, _43 * fInv, _44 * fInv);
}

inline XMATRIX operator * ( float f, const XMATRIX& mat )
{
    return XMATRIX(f * mat._11, f * mat._12, f * mat._13, f * mat._14,
                   f * mat._21, f * mat._22, f * mat._23, f * mat._24,
                   f * mat._31, f * mat._32, f * mat._33, f * mat._34,
                   f * mat._41, f * mat._42, f * mat._43, f * mat._44);
}

inline bool XMATRIX::operator == ( const XMATRIX& mat ) const
{
    return 0 == memcmp(this, &mat, sizeof(XMATRIX));
}

inline bool XMATRIX::operator != ( const XMATRIX& mat ) const
{
    return 0 != memcmp(this, &mat, sizeof(XMATRIX));
}

/////////////////////////////////////////////////////////////////////////////
// Quaternion
/////////////////////////////////////////////////////////////////////////////

inline QUATERNION::QUATERNION( const float* pf )
{
	x = pf[0];
	y = pf[1];
	z = pf[2];
	w = pf[3];
}

inline QUATERNION::QUATERNION( float fx, float fy, float fz, float fw )
{
	x = fx;
	y = fy;
	z = fz;
	w = fw;
}
// casting
inline QUATERNION::operator float* ()
{
    return (float *) &x;
}

inline QUATERNION::operator const float* () const
{
    return (const float *) &x;
}


// assignment operators
inline QUATERNION& QUATERNION::operator += ( const QUATERNION& q )
{
	x += q.x;
	y += q.y;
	z += q.z;
	w += q.w;
	return *this;
}

inline QUATERNION& QUATERNION::operator -= ( const QUATERNION& q )
{
	x -= q.x;
	y -= q.y;
	z -= q.z;
	w -= q.w;
	return *this;
}

inline QUATERNION& QUATERNION::operator *= ( const QUATERNION& q )
{
    QuaternionMultiply(this, this, &q);
    return *this;
}

inline QUATERNION& QUATERNION::operator *= ( float f )
{
	x *= f;
	y *= f;
	z *= f;
	w *= f;
	return *this;
}

inline QUATERNION& QUATERNION::operator /= ( float f )
{
	float fInv = 1.0f / f;
	x *= fInv;
	y *= fInv;
	z *= fInv;
	w *= fInv;
	return *this;
}

// unary operators
inline QUATERNION QUATERNION::operator + () const
{
	return *this;
}

inline QUATERNION QUATERNION::operator - () const
{
	return QUATERNION(-x, -y, -z, -w);
}

// binary operators
inline QUATERNION QUATERNION::operator + ( const QUATERNION& q ) const
{
	return QUATERNION(x + q.x, y + q.y, z + q.z, w + q.w);
}

inline QUATERNION QUATERNION::operator - ( const QUATERNION& q ) const
{
	return QUATERNION(x - q.x, y - q.y, z - q.z, w - q.w);
}

inline QUATERNION QUATERNION::operator * ( const QUATERNION& q ) const
{
	QUATERNION qT;
	QuaternionMultiply(&qT, this, &q);
	return qT;
}

inline QUATERNION QUATERNION::operator * ( float f ) const
{
	return QUATERNION(x * f, y * f, z * f, w * f);
}

inline QUATERNION QUATERNION::operator / ( float f ) const
{
	float fInv = 1.0f / f;
	return QUATERNION(x * fInv, y * fInv, z * fInv, w * fInv);
}

inline QUATERNION operator * (float f, const QUATERNION& q )
{
	return QUATERNION(f * q.x, f * q.y, f * q.z, f * q.w);
}


inline bool QUATERNION::operator == ( const QUATERNION& q ) const
{
	return x == q.x && y == q.y && z == q.z && w == q.w;
}

inline bool QUATERNION::operator != ( const QUATERNION& q ) const
{
	return x != q.x || y != q.y || z != q.z || w != q.w;
}

/////////////////////////////////////////////////////////////////////////////
// Inline functions
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// 3D Vector
/////////////////////////////////////////////////////////////////////////////

inline float Vec3Length( const VECTOR3 *pV )
{
	return sqrtf(pV->x * pV->x + pV->y * pV->y + pV->z * pV->z);
}

inline float Vec3LengthSq( const VECTOR3 *pV )
{
	return pV->x * pV->x + pV->y * pV->y + pV->z * pV->z;
}

inline float Vec3Dot( const VECTOR3 *pV1, const VECTOR3 *pV2 )
{
	return pV1->x * pV2->x + pV1->y * pV2->y + pV1->z * pV2->z;
}

inline VECTOR3* Vec3Cross( VECTOR3 *pOut, const VECTOR3 *pV1, const VECTOR3 *pV2 )
{
	VECTOR3 v;

	v.x = pV1->y * pV2->z - pV1->z * pV2->y;
	v.y = pV1->z * pV2->x - pV1->x * pV2->z;
	v.z = pV1->x * pV2->y - pV1->y * pV2->x;

	*pOut = v;
	return pOut;
}

inline VECTOR3* Vec3Add( VECTOR3 *pOut, const VECTOR3 *pV1, const VECTOR3 *pV2 )
{
	pOut->x = pV1->x + pV2->x;
	pOut->y = pV1->y + pV2->y;
	pOut->z = pV1->z + pV2->z;
	return pOut;
}

inline VECTOR3* Vec3Subtract( VECTOR3 *pOut, const VECTOR3 *pV1, const VECTOR3 *pV2 )
{
	pOut->x = pV1->x - pV2->x;
	pOut->y = pV1->y - pV2->y;
	pOut->z = pV1->z - pV2->z;
	return pOut;
}

inline VECTOR3* Vec3Minimize( VECTOR3 *pOut, const VECTOR3 *pV1, const VECTOR3 *pV2 )
{
	pOut->x = pV1->x < pV2->x ? pV1->x : pV2->x;
	pOut->y = pV1->y < pV2->y ? pV1->y : pV2->y;
	pOut->z = pV1->z < pV2->z ? pV1->z : pV2->z;
	return pOut;
}

inline VECTOR3* Vec3Maximize( VECTOR3 *pOut, const VECTOR3 *pV1, const VECTOR3 *pV2 )
{
	pOut->x = pV1->x > pV2->x ? pV1->x : pV2->x;
	pOut->y = pV1->y > pV2->y ? pV1->y : pV2->y;
	pOut->z = pV1->z > pV2->z ? pV1->z : pV2->z;
	return pOut;
}

inline VECTOR3* Vec3Scale( VECTOR3 *pOut, const VECTOR3 *pV, float s)
{
	pOut->x = pV->x * s;
	pOut->y = pV->y * s;
	pOut->z = pV->z * s;
	return pOut;
}

inline VECTOR3* Vec3Lerp( VECTOR3 *pOut, const VECTOR3 *pV1, const VECTOR3 *pV2, float s )
{
	pOut->x = pV1->x + s * (pV2->x - pV1->x);
	pOut->y = pV1->y + s * (pV2->y - pV1->y);
	pOut->z = pV1->z + s * (pV2->z - pV1->z);
	return pOut;
}

/////////////////////////////////////////////////////////////////////////////
// 4D Matrix
/////////////////////////////////////////////////////////////////////////////

inline XMATRIX* MatrixIdentity( XMATRIX *pOut )
{
	pOut->m[0][1] = pOut->m[0][2] = pOut->m[0][3] =
	pOut->m[1][0] = pOut->m[1][2] = pOut->m[1][3] =
	pOut->m[2][0] = pOut->m[2][1] = pOut->m[2][3] =
	pOut->m[3][0] = pOut->m[3][1] = pOut->m[3][2] = 0.0f;

	pOut->m[0][0] = pOut->m[1][1] = pOut->m[2][2] = pOut->m[3][3] = 1.0f;
	return pOut;
}

inline bool MatrixIsIdentity( const XMATRIX *pM )
{
	return pM->m[0][0] == 1.0f && pM->m[0][1] == 0.0f && pM->m[0][2] == 0.0f && pM->m[0][3] == 0.0f &&
           pM->m[1][0] == 0.0f && pM->m[1][1] == 1.0f && pM->m[1][2] == 0.0f && pM->m[1][3] == 0.0f &&
           pM->m[2][0] == 0.0f && pM->m[2][1] == 0.0f && pM->m[2][2] == 1.0f && pM->m[2][3] == 0.0f &&
           pM->m[3][0] == 0.0f && pM->m[3][1] == 0.0f && pM->m[3][2] == 0.0f && pM->m[3][3] == 1.0f;
}

// Build a matrix which translates by (x, y, z)
inline XMATRIX* MatrixTranslation ( XMATRIX *pOut, float x, float y, float z )
{

    MatrixIdentity(pOut);
    pOut[3][0] = x;
    pOut[3][1] = y;
    pOut[3][2] = z;

	return pOut;
}

// Build a rotation matrix by X axis
inline XMATRIX* MatrixRotationX ( XMATRIX *pOut, float angle )
{
	float fSin = (float)sin(angle), 
		  fCos = (float)cos(angle);

    pOut->_11 =  1.0f; pOut->_12 =  0.0f;   pOut->_13 =  0.0f; pOut->_14 =  0.0f;
    pOut->_21 =  0.0f; pOut->_22 =  fCos;   pOut->_23 =  fSin; pOut->_24 =  0.0f;
    pOut->_31 =  0.0f; pOut->_32 = -fSin;   pOut->_33 =  fCos; pOut->_34 =  0.0f;
    pOut->_41 =  0.0f; pOut->_42 =  0.0f;   pOut->_43 =  0.0f; pOut->_44 =  1.0f;

    return pOut;
}

// Build a rotation matrix by Y axis
inline XMATRIX* MatrixRotationY ( XMATRIX *pOut, float angle )
{
	float fSin = (float)sin(angle), 
		  fCos = (float)cos(angle);

    pOut->_11 =  fCos; pOut->_12 =  0.0f;   pOut->_13 = -fSin; pOut->_14 =  0.0f;
    pOut->_21 =  0.0f; pOut->_22 =  1.0f;   pOut->_23 =  0.0f; pOut->_24 =  0.0f;
    pOut->_31 =  fSin; pOut->_32 =  0.0f;   pOut->_33 =  fCos; pOut->_34 =  0.0f;
    pOut->_41 =  0.0f; pOut->_42 =  0.0f;   pOut->_43 =  0.0f; pOut->_44 =  1.0f;

    return pOut;
}

// Build a rotation matrix by Z axis
inline XMATRIX* MatrixRotationZ ( XMATRIX *pOut, float angle )
{
	float fSin = (float)sin(angle), 
		  fCos = (float)cos(angle);

    pOut->_11 =  fCos; pOut->_12 =  fSin;   pOut->_13 =  0.0f; pOut->_14 =  0.0f;
    pOut->_21 = -fSin; pOut->_22 =  fCos;   pOut->_23 =  0.0f; pOut->_24 =  0.0f;
    pOut->_31 =  0.0f; pOut->_32 =  0.0f;   pOut->_33 =  1.0f; pOut->_34 =  0.0f;
    pOut->_41 =  0.0f; pOut->_42 =  0.0f;   pOut->_43 =  0.0f; pOut->_44 =  1.0f;

    return pOut;
}

inline VECTOR3 TransformVector (XMATRIX *pM1, VECTOR3 *pV1 )
{
	VECTOR3 vRes;

	vRes.x = pM1->_11*pV1->x + pM1->_12*pV1->y + pM1->_13*pV1->z + pM1->_14;
	vRes.y = pM1->_21*pV1->x + pM1->_22*pV1->y + pM1->_23*pV1->z + pM1->_24;
	vRes.z = pM1->_31*pV1->x + pM1->_32*pV1->y + pM1->_33*pV1->z + pM1->_34;

	return vRes;
}

/////////////////////////////////////////////////////////////////////////////
// Quaternion
/////////////////////////////////////////////////////////////////////////////

inline float QuaternionLength( const QUATERNION *pQ )
{
	return sqrtf(pQ->x * pQ->x + pQ->y * pQ->y + pQ->z * pQ->z + pQ->w * pQ->w);
}

inline float QuaternionLengthSq( const QUATERNION *pQ )
{
	return pQ->x * pQ->x + pQ->y * pQ->y + pQ->z * pQ->z + pQ->w * pQ->w;
}

inline float QuaternionDot( const QUATERNION *pQ1, const QUATERNION *pQ2 )
{
	return pQ1->x * pQ2->x + pQ1->y * pQ2->y + pQ1->z * pQ2->z + pQ1->w * pQ2->w;
}

inline QUATERNION* QuaternionIdentity( QUATERNION *pOut )
{
	pOut->x = pOut->y = pOut->z = 0.0f;
	pOut->w = 1.0f;
	return pOut;
}

inline bool QuaternionIsIdentity( const QUATERNION *pQ )
{
	return pQ->x == 0.0f && pQ->y == 0.0f && pQ->z == 0.0f && pQ->w == 1.0f;
}

inline QUATERNION* QuaternionConjugate( QUATERNION *pOut, const QUATERNION *pQ )
{
	pOut->x = -pQ->x;
	pOut->y = -pQ->y;
	pOut->z = -pQ->z;
	pOut->w =  pQ->w;
	return pOut;
}

#endif