//***************************************************************************
//
// COMMON.H
// by Harlen C. Batagelo
//
//***************************************************************************

#ifndef _COMMON_H
#define _COMMON_H

#include "fps.h"
#include "arcball.h"
#include "bitset.h"

#include <GL/glut.h>

class CDisk;

// STL
#pragma warning(disable:4786)
#include <set>
#include <list>

typedef std::set<int> set_of_ids;

// common definitions

#define USETBV

#ifndef FLT_MAX
#define FLT_MAX 3.402823466e+38F // max float value
#endif

#define SIZE_X  70 // data structure resolution
#define SIZE_Y  70
#define SIZE_Z  70

#define SIZE_XM1 69 // data structure resolution (minus 1)
#define SIZE_YM1 69
#define SIZE_ZM1 69

#define SIZE_XY  4900 // SIZE_X*SIZE_Y

#define WIN_WIDTH  600	// window resolution
#define WIN_HEIGHT 300						 

#define N_OBJS_DYN 200 // dynamic objects
#define N_OBJS_STC 4800 // static objects

#define ROUND(x) ((int)((x)+((x<0)?-.5:.5)))
#define MAX(a,b) ((a>b)?(a):(b))
#define MIN(a,b) ((a<b)?(a):(b))
#define ISINSIDE(x,y,z,r) ((r)*(r)>=(x)*(x)+(y)*(y)+(z)*(z)) // r^2 >= x^2+y^2+z^2

#define SQUARE3(a,b,c) (sqrt((a)*(a)+(b)*(b)+(c)*(c)))
#define RAD(x) ((x) * 1.74532925199e-2)

#define STRUCTURE_DISPLAY_LIST 1 // display list ID

#define STEP 25.0 / max(max(SIZE_X, SIZE_Y), SIZE_Z)

#define MK_OFF(y,x,w) (((y)*(w))+(x)) // make offset
#define MK_OFF_GRID(x,y,z) (((z)*SIZE_XY)+((y)*SIZE_X)+(x)) // make offset for O,H,etc

#define isSeedOutsideBnd(x,y,z) ((((x)>=0 && (x)<SIZE_X) && ((y)>=0 && (y)<SIZE_Y) && ((z)>=0 && (z)<SIZE_Z)) ? 0 : 1)
#define isInsideVF(x,y,z) (((a_right*(x)+b_right*(y)+c_right*(z)+d_right<0) && (a_left*(x)+b_left*(y)+c_left*(z)+d_left<0) && (a_bottom*(x)+b_bottom*(y)+c_bottom*(z)+d_bottom<0) && (a_top*(x)+b_top*(y)+c_top*(z)+d_top<0) && (a_near*(x)+b_near*(y)+c_near*(z)+d_near<0)) ? 1 : 0)

// Definition of a 2D vertex
class VERTEX2D {
public:
	float fX, fY, fIncX, fIncY;

	inline void Set (float x, float y) { fX = x; fY = y; }
	inline void SetInc (float x, float y) { fIncX = x; fIncY = y; }
	inline void Increment (void) { fX += fIncX; fY += fIncY; }
};

// Definition of a bucket of edges
class CONCPOLLINE {
public:
	int iEdges;
	int iIdx1[256], iIdx2[256];

	CONCPOLLINE() { iEdges = 0; }
	inline void Reset (void) { iEdges = 0; }
	inline void Set1stVertex (int iIdx) { iIdx1[iEdges] = iIdx; }
	inline void Set2ndVertex (int iIdx) { iIdx2[iEdges] = iIdx; }
	inline void NewEdge (void) { iEdges++; }
};

// Definition of a bucket of ending vertices
class CONCPOLEND {
public:
	int iElements;
	int iIdxEnd[256];

	CONCPOLEND() { iElements = 0; }
	inline void Reset (void) { iElements = 0; }
	inline void Add (int iIdx) { iIdxEnd[iElements++] = iIdx; }
};

enum rasterPlane { raFront, raBack, raLeft, raRight, raTop, raDown };

class Vis3D
{
	int current_date;
	float fps_avg;
	int fps_avg_ct;

	int iMaxDist; // maximum depth of a shadow volume

	// clipping planes coefficients
	float a_left, b_left, c_left, d_left;
	float a_right, b_right, c_right, d_right;
	float a_top, b_top, c_top, d_top;
	float a_bottom, b_bottom, c_bottom, d_bottom;
	float a_near, b_near, c_near, d_near;

	VECTOR3 tCorner1, tCorner2, tCorner3, tCorner4; // corner points that define the view frustum (transformed)
	float zFar; // distance of the far plane from the viewpoint

	// regular grid structures
	BitSet H, O, Os,         // H, O, Os
		   Txy[N_OBJS_DYN],  // T projection planes
		   Tyz[N_OBJS_DYN],
		   Txz[N_OBJS_DYN];
	set_of_ids *TxyN, *TyzN, *TxzN;    // Number of elements of Txy, Tyz and Txz
	set_of_ids *I; // I

	bool *PVS;      // PVS (each entrie indicates the state of a particular object)
	bool *PVSVF;    // PVS for view-frustum culling only
	bool *PVSEXACT; // PVS for exact visibility calculation

	unsigned char *ucReadBufferR, // Used for exact visibility calculation
		  *ucReadBufferG,
		  *ucReadBufferB;

	CDisk *disks; // dynamic/static objects

	// used to store a color for each object
	float R[N_OBJS_DYN+N_OBJS_STC], 
		  G[N_OBJS_DYN+N_OBJS_STC], 
		  B[N_OBJS_DYN+N_OBJS_STC];

public:
	GLUquadricObj *qobj; // used to draw disks, cylinders, etc

	bool fusionEnabled; // true for occluder fusion
	bool hiddenGeometry; // true for scene visualization
	bool bCalcExactVis;  // true for exact visibility calculation
	bool bDynamicScene;  // true for moving objects

	// raster2vector matrices
	int *iMatrixF, *iMatrixB, // front, back
	    *iMatrixL, *iMatrixR, // left, right
		*iMatrixT, *iMatrixD; // top, down
	char bMatrixF, bMatrixB, // flags to indicate whether the corresponding raster matrix is valid
		 bMatrixL, bMatrixR,
		 bMatrixT, bMatrixD;

	float fCPMinY, fCPMaxY, fCPMinX, fCPMaxX; // min/max coordinates of the current raster image

	int iCPMinXF, iCPMinXB, iCPMinXL, iCPMinXR, iCPMinXT, iCPMinXD,
		iCPMinYF, iCPMinYB, iCPMinYL, iCPMinYR, iCPMinYT, iCPMinYD,
		iCPMaxXF, iCPMaxXB, iCPMaxXL, iCPMaxXR, iCPMaxXT, iCPMaxXD,
		iCPMaxYF, iCPMaxYB, iCPMaxYL, iCPMaxYR, iCPMaxYT, iCPMaxYD;

	// bucket of vertices per line for concave polygon rasterization
	CONCPOLLINE cpVLine[512]; 
	// indices of ending vertices of perpendicular spans	
	CONCPOLEND cpVLineEnd[512];

	CONCPOLLINE cpHLine[512]; 
	CONCPOLEND cpHLineEnd[512];

	int iVertices, iVLines, iHLines;
	VERTEX2D vtVertices[16384];

	bool bRasterFull; // used to terminate the traversal prematurely
	bool bRasterFullF, bRasterFullB, bRasterFullL, bRasterFullR, bRasterFullT, bRasterFullD;
	bool bPrematureEnd;

	int iDist, // chess distance from the viewpoint to the seed-cell
		iDistSize; // size (width or height) of a plane where in which every cell has the same chess distance, i.e., iDist

	int iRasterPlane;		// used to store values of 'rasterPlane' enum

	VECTOR3 vPos;			// viewpoint position
	int iOrX, iOrY, iOrZ;	// viewpoint position in integer coordinates
	float rotX, rotY, rotZ; // rotation angle (in degrees)
	float fFOV;				// field of view (in degrees)
	XMATRIX mView;			// view matrix
	CArcBall arcBall;		// arcball class

	BOOL bShowData;

	Vis3D ();
	~Vis3D ();

	inline bool setO (int iX, int iY, int iZ)
	{
		if (iX < 0 || iX > SIZE_XM1) return false;
		if (iY < 0 || iY > SIZE_YM1) return false;
		if (iZ < 0 || iZ > SIZE_ZM1) return false;
		O.set(MK_OFF_GRID(iX,iY,iZ));
		return true;
	}

	inline bool setOs (int iX, int iY, int iZ)
	{
		if (iX < 0 || iX > SIZE_XM1) return false;
		if (iY < 0 || iY > SIZE_YM1) return false;
		if (iZ < 0 || iZ > SIZE_ZM1) return false;
		Os.set(MK_OFF_GRID(iX,iY,iZ));
		return true;
	}

	inline bool setI (int iX, int iY, int iZ, int iID)
	{
		if (iX < 0 || iX > SIZE_XM1) return false;
		if (iY < 0 || iY > SIZE_YM1) return false;
		if (iZ < 0 || iZ > SIZE_ZM1) return false;
		I[MK_OFF_GRID(iX,iY,iZ)].insert (iID);
		return true;
	}

	inline bool setTxy (int iX, int iY, int iID)
	{
		if (iX < 0 || iX > SIZE_XM1) return false;
		if (iY < 0 || iY > SIZE_YM1) return false;
		Txy[iID].set(MK_OFF(iY,iX,SIZE_X));
		TxyN[MK_OFF(iY,iX,SIZE_X)].insert (iID);
		return true;
	}

	inline bool setLineTxy (int iX1, int iX2, int iY, int iID)
	{
		if (iY < 0 || iY > SIZE_YM1) return false;
		if (iX1 < 0) iX1 = 0; if (iX2 > SIZE_XM1) iX2 = SIZE_XM1;
		if (iX2 > iX1)
			Txy[iID].set(MK_OFF(iY,iX1,SIZE_X), (iX2-iX1)+1);

		for (int i = iX1; i <= iX2; i++)
			TxyN[MK_OFF(iY,i,SIZE_X)].insert (iID);

		return true;
	}

	inline bool setTyz (int iZ, int iY, int iID)
	{
		if (iY < 0 || iY > SIZE_YM1) return false;
		if (iZ < 0 || iZ > SIZE_ZM1) return false;
		Tyz[iID].set(MK_OFF(iY,iZ,SIZE_Z));
		TyzN[MK_OFF(iY,iZ,SIZE_Z)].insert (iID);
		return true;
	}

	inline bool setLineTyz (int iZ1, int iZ2, int iY, int iID)
	{
		if (iY < 0 || iY > SIZE_YM1) return false;
		if (iZ1 < 0) iZ1 = 0; if (iZ2 > SIZE_ZM1) iZ2 = SIZE_ZM1;
		if (iZ2 > iZ1)
			Tyz[iID].set(MK_OFF(iY,iZ1,SIZE_Z), (iZ2-iZ1)+1);

		for (int i = iZ1; i <= iZ2; i++)
			TyzN[MK_OFF(iY,i,SIZE_Z)].insert (iID);

		return true;
	}

	inline bool setTxz (int iX, int iZ, int iID)
	{
		if (iX < 0 || iX > SIZE_XM1) return false;
		if (iZ < 0 || iZ > SIZE_ZM1) return false;
		Txz[iID].set(MK_OFF(iZ,iX,SIZE_X));
		TxzN[MK_OFF(iZ,iX,SIZE_X)].insert (iID);
		return true;
	}

	inline bool setLineTxz (int iX1, int iX2, int iZ, int iID)
	{
		if (iZ < 0 || iZ > SIZE_ZM1) return false;
		if (iX1 < 0) iX1 = 0; if (iX2 > SIZE_XM1) iX2 = SIZE_XM1;
		if (iX2 > iX1)
			Txz[iID].set(MK_OFF(iZ,iX1,SIZE_X), (iX2-iX1)+1);

		for (int i = iX1; i <= iX2; i++)
			TxzN[MK_OFF(iZ,i,SIZE_X)].insert (iID);

		return true;
	}

	inline bool resetTxy (int iX, int iY, int iID)
	{
		if (iX < 0 || iX > SIZE_XM1) return false;
		if (iY < 0 || iY > SIZE_YM1) return false;
		Txy[iID].reset(MK_OFF(iY,iX,SIZE_X));
		TxyN[MK_OFF(iY,iX,SIZE_X)].erase (iID);
		return true;
	}

	inline bool resetLineTxy (int iX1, int iX2, int iY, int iID)
	{
		if (iY < 0 || iY > SIZE_YM1) return false;
		if (iX1 < 0) iX1 = 0; if (iX2 > SIZE_XM1) iX2 = SIZE_XM1;
		if (iX2 > iX1)
			Txy[iID].reset(MK_OFF(iY,iX1,SIZE_X), (iX2-iX1)+1);

		for (int i = iX1; i <= iX2; i++)
			TxyN[MK_OFF(iY,i,SIZE_X)].erase (iID);

		return true;
	}

	inline bool resetTyz (int iZ, int iY, int iID)
	{
		if (iY < 0 || iY > SIZE_YM1) return false;
		if (iZ < 0 || iZ > SIZE_ZM1) return false;
		Tyz[iID].reset(MK_OFF(iY,iZ,SIZE_Z));
		TyzN[MK_OFF(iY,iZ,SIZE_Z)].erase (iID);
		return true;
	}

	inline bool resetLineTyz (int iZ1, int iZ2, int iY, int iID)
	{
		if (iY < 0 || iY > SIZE_YM1) return false;
		if (iZ1 < 0) iZ1 = 0; if (iZ2 > SIZE_ZM1) iZ2 = SIZE_ZM1;
		if (iZ2 > iZ1)
			Tyz[iID].reset(MK_OFF(iY,iZ1,SIZE_Z), (iZ2-iZ1)+1);

		for (int i = iZ1; i <= iZ2; i++)
			TyzN[MK_OFF(iY,i,SIZE_Z)].erase (iID);

		return true;
	}

	inline bool resetTxz (int iX, int iZ, int iID)
	{
		if (iX < 0 || iX > SIZE_XM1) return false;
		if (iZ < 0 || iZ > SIZE_ZM1) return false;
		Txz[iID].reset(MK_OFF(iZ,iX,SIZE_X));
		TxzN[MK_OFF(iZ,iX,SIZE_X)].erase (iID);
		return true;
	}

	inline bool resetLineTxz (int iX1, int iX2, int iZ, int iID)
	{
		if (iZ < 0 || iZ > SIZE_ZM1) return false;
		if (iX1 < 0) iX1 = 0; if (iX2 > SIZE_XM1) iX2 = SIZE_XM1;
		if (iX2 > iX1)
			Txz[iID].reset(MK_OFF(iZ,iX1,SIZE_X), (iX2-iX1)+1);

		for (int i = iX1; i <= iX2; i++)
			TxzN[MK_OFF(iZ,i,SIZE_X)].erase (iID);

		return true;
	}

	// MAIN.CPP
	void putIntoGrid	 (int iCt, bool bStatic, bool bEraseOnly);
	void init            (void);
	void startTraversal  (int iX, int iY, int iZ);
	void bresenhamLine3D (int x1, int y1, int z1, int x2, int y2, int z2);
	void show            (void);

	// GEOPRI.CPP
	inline float fTheta (int x1, int y1, int x2, int y2)
	{
		int dx = x2 - x1, dy = y2 - y1;
		float t = (!dx && !dy) ? -99.0 : (float) dy / (abs(dx) + abs(dy)); 
		if (dx < 0) { t = 2 - t; } else { if (dy < 0) t = 4 + t; }

		return t;
	}

	int isInsideVFBnd              (float x, float y, float z, VECTOR3 &vBound, int &iOFB, int &iXOFB, int &iYOFB, int &iZOFB);
	VECTOR3 rayFrustumIntersect    (VECTOR3 &vOrg, VECTOR3 &vDir);
	VECTOR3 rayFrustumIntersectBnd (VECTOR3 &vOrg, VECTOR3 &vDir, VECTOR3 &vBound, int &iOFB, int &iXOFB, int &iYOFB, int &iZOFB);
	VECTOR3 getNormal              (VECTOR3 &p1, VECTOR3 &p2, VECTOR3 &p3);
	void transformVFC			   (void);

	void ras2Vec				   (int *iRas, int iSize);
	void drawConcPolF			   (void);
	void drawConcPolB			   (void);
	void drawConcPolL			   (void);
	void drawConcPolR			   (void);
	void drawConcPolT			   (void);
	void drawConcPolD			   (void);

	// TRAVERSAL.CPP

	void insertTBV (int disk);
	void eraseTBV (int disk);
	void Add2PVS (int iX, int iY, int iZ);
	void TestCell (int iX, int iY, int iZ);
//	void TestCellFootprint (int iX, int iY, int iZ);

	void YXTraversalNeg (int iX, int iY, int iZ, int iSeedXOFB, int iSeedYOFB);
	void Y1TraversalNeg (int iX, int iY, int iZ, int iSeedOFB);
	void YXTraversalPos (int iX, int iY, int iZ, int iSeedXOFB, int iSeedYOFB);
	void Y1TraversalPos (int iX, int iY, int iZ, int iSeedOFB);

	void YZTraversalNeg (int iX, int iY, int iZ, int iSeedYOFB, int iSeedZOFB);
	void Y2TraversalNeg (int iX, int iY, int iZ, int iSeedOFB);
	void YZTraversalPos (int iX, int iY, int iZ, int iSeedYOFB, int iSeedZOFB);
	void Y2TraversalPos (int iX, int iY, int iZ, int iSeedOFB);

	void XYTraversalNeg (int iX, int iY, int iZ, int iSeedXOFB, int iSeedYOFB);
	void XTraversalNeg  (int iX, int iY, int iZ, int iSeedOFB);
	void XYTraversalPos (int iX, int iY, int iZ, int iSeedXOFB, int iSeedYOFB);
	void XTraversalPos  (int iX, int iY, int iZ, int iSeedOFB);

	void ZXTraversalNeg (int iX, int iY, int iZ, int iSeedOFB);
	void Z1TraversalNeg (int iX, int iY, int iZ, int iSeedOFB);
	void ZXTraversalPos (int iX, int iY, int iZ, int iSeedOFB);
	void Z1TraversalPos (int iX, int iY, int iZ, int iSeedOFB);

	// special 'diagonal' cases
	void YXTraversalNegDg (int iX, int iY, int iZ, int iSeedOFB);
	void Y1TraversalNegDg (int iX, int iY, int iZ, int iSeedOFB);
	void YXTraversalPosDg (int iX, int iY, int iZ, int iSeedOFB);
	void Y1TraversalPosDg (int iX, int iY, int iZ, int iSeedOFB);

	void XYTraversalNegDg (int iX, int iY, int iZ, int iSeedOFB);
	void XTraversalNegDg  (int iX, int iY, int iZ, int iSeedOFB);
	void XYTraversalPosDg (int iX, int iY, int iZ, int iSeedOFB);
	void XTraversalPosDg  (int iX, int iY, int iZ, int iSeedOFB);

	// TEXT.CPP
	void printString (float x, float y, char *text, ...);
	void drawCell    (int x, int y, int z);
};

#endif